#include "BSTree.h"
#include <ostream>

BSTree::~BSTree ()
{
   while (!empty())
   {
      delete remove(root);
   }
}


BSTree::BSTNode* 
BSTree::BSTNode::min(BSTree::BSTNode* nil)
{
   BSTNode* p = this;
   while (p->left != nil)
   {
      p = p->left;
   }
   return p;
}


BSTree::BSTNode* 
BSTree::BSTNode::max(BSTree::BSTNode* nil)
{
   BSTNode* p = this;
   while (p->right != nil)
   {
      p = p->right;
   }
   return p;
}


BSTree::BSTNode* 
BSTree::BSTNode::find(KEY const& k, BSTree::BSTNode* nil)
{
   BSTNode* p = this;
   while (p != nil && k != p->key)
   {
      if (k < p->key)
         p = p->left;
      else
         p = p->right;
   }
   return p;
}


BSTree::BSTNode* 
BSTree::BSTNode::next(BSTNode* nil)
{
   BSTNode* p = this;
   assert (p != nil);
   if (right != nil)
      return right->min(nil);
   BSTNode* y = p->parent;
   while( y != nil && p == y->right)
   {
      p = y;
      y = y->parent;
   }
   return y;
}


void 
BSTree::BSTNode::print(std::ostream& F,  BSTNode* nil)
{
   const char c = '.';
   if (this == nil)
      return;
   if (left!= nil)
   {
      F << "[";
      this->left->print(F, nil);
      F << ']';
   }
   else F << c;
   F << key;
   if (right != nil)
   {
      F << '(';
      this->right->print(F, nil);
      F << ')';
   }
   else
      F << c ;
}


void BSTree::print(std::ostream& F)
{
   if (!empty())
      root->print(F,  nil);
   else
      F << "<pusty>";
   F << "\t s: " << size() << ",\t h: " << height() << "\n";
}


BSTree::BSTNode* 
BSTree::find(KEY const& key)
{
   if (empty())
      return root;
   return root->find(key, this->nil);
}


void 
BSTree::insert(BSTNode* node)
{
   _size++;
   BSTNode* y = nil;
   BSTNode* x = root;
   while (x != nil)
   {
      y = x;
      if (node->key < x->key)
         x = x->left;
      else
         x = x->right;
   }
   node->parent = y;
   if (y == nil)
      root = node;
   else
      if (node->key < y->key)
         y->left = node;
      else
         y->right = node;
}


void 
BSTree::insert(KEY const& key)
{
   BSTNode* p = new BSTNode (key, nil);
   insert(p);
}


BSTree::BSTNode*
BSTree::remove(BSTNode* node)
{
   if (node == nil) 
      return nil;

   _size--;

   BSTNode* usuniety = (node->left == nil || node->right == nil) ?
                       node  :  node->next(nil);
   BSTNode* galaz = (usuniety->left != nil) ?
                 galaz = usuniety->left  :  usuniety->right;

   if (galaz != nil)
      galaz->parent = usuniety->parent;
   if (usuniety->parent == nil)
      root = galaz;
   else if (usuniety == usuniety->parent->left)
      usuniety->parent->left = galaz;
   else
      usuniety->parent->right = galaz;
   if (usuniety != node)
      node->key = usuniety->key;
   return usuniety;
}

unsigned int BSTree::BSTNode::height(BSTree::BSTNode* nil) const
{
   if (this == nil) 
      return 0;
   unsigned int x = left->height(nil);
   unsigned int y = right->height(nil);
   return  (x > y) ? x + 1 : y + 1;
}