#ifndef BSTREE__H__
#define BSTREE__H__

#include <cassert>
#include <ostream>

typedef int KEY;

class BSTree
{
public:
   struct BSTNode
   {
      BSTNode* parent;
      BSTNode* right;
      BSTNode* left;
      KEY key;

      BSTNode(KEY const& k, BSTNode* p, BSTNode* r, BSTNode* l);
      BSTNode(KEY const& k, BSTNode* nil);

      BSTNode* find(KEY const& key, BSTNode* nil);
      BSTNode* min(BSTNode* nil);
      BSTNode* max(BSTNode* nil);
      BSTNode* next(BSTNode* nil);
      unsigned int height(BSTNode* nil) const; 

      void print(std::ostream& F, BSTNode* nil);
   };

   BSTree();
   ~BSTree();

   void insert(KEY const& key);
   void remove(KEY const& key);
   BSTNode* find(KEY const& key);
  
   void print(std::ostream& F);
   bool empty() const                              { return root == nil; }
   unsigned int size() const                       { return _size; }
   unsigned int height() const                    { return root->height(nil); }

private:
   void insert (BSTNode* pnode);
   BSTree::BSTNode* remove (BSTNode* pnode);

   BSTNode* root;
   BSTNode  _nil;
   BSTNode* nil;
   unsigned int _size;
};



// INLINE

inline
BSTree::BSTNode::BSTNode(KEY const& k, BSTNode* p, BSTNode* r, BSTNode* l)
: parent(p), right(r), left(l), key(k)
{ }

inline
BSTree::BSTNode::BSTNode(KEY const& k, BSTNode* nil)
: parent(nil), right(nil), left(nil), key(k)
{ }

inline
BSTree::BSTree()
: root(&_nil), _nil(0, &_nil, &_nil, &_nil), nil(&_nil), _size(0)
{ }

inline void 
BSTree::remove(KEY const& key)
{
   remove(find(key));
}



#endif