#include "RB_BSTree.h"
#include <iostream>

void INFO(char const* s)
{
#ifndef NDEBUG
   std::cerr << s;
#endif
}


RBTree::~RBTree ()
{
   while (!empty())
   {
      delete remove(root);
   }
}


RBTree::RBNode* 
RBTree::RBNode::min(RBTree::RBNode* nil)
{
   RBNode* p = this;
   while (p->left != nil)
   {
      p = p->left;
   }
   return p;
}


RBTree::RBNode* 
RBTree::RBNode::max(RBTree::RBNode* nil)
{
   RBNode* p = this;
   while (p->right != nil)
   {
      p = p->right;
   }
   return p;
}


RBTree::RBNode* 
RBTree::RBNode::find(KEY const& k, RBTree::RBNode* nil)
{
   RBNode* p = this;
   while (p != nil && k != p->key)
   {
      if (k < p->key)
         p = p->left;
      else
         p = p->right;
   }
   return p;
}


RBTree::RBNode* 
RBTree::RBNode::next(RBNode* nil)
{
   RBNode* p = this;
   assert (p != nil);
   if (right != nil)
      return right->min(nil);
   RBNode* y = p->parent;
   while( y != nil && p == y->right)
   {
      p = y;
      y = y->parent;
   }
   return y;
}


void 
RBTree::RBNode::print(std::ostream& F,  RBNode* nil)
{
 
   const char c = '.';
   if (this == nil)
      return;
   if (left != nil)
   {
      F << '[';
      this->left->print(F, nil);
      F << ']';
   }
   else F << c;
   if (color == Red) 
      F << 'R';
   F << key;
   if (right != nil)
   {
      F << '(';
      this->right->print(F, nil);
      F << ')';
   }
   else
      F << c ;
}


void 
RBTree::RBNode::check_it(RBNode* nil)
{
   assert (color == Black || color == Red);
   assert (color == Black || left->color == Black || right->color == Black);
   assert (left->black_height(nil) == right->black_height(nil));
   assert (left == nil  || left->key <= key);
   assert (right == nil || right->key >= key);

}


void RBTree::print(std::ostream& F)
{
   if (!empty())
      root->print(F,  nil);
   else
      F << "<pusty>";
   F << "\t s: " << size() << ",\t h: " << height() << "\n";
}


RBTree::RBNode* 
RBTree::find(KEY const& key)
{
   if (empty())
      return root;
   return root->find(key, this->nil);
}


void 
RBTree::insert(RBNode* node)
{
   _size++;
   RBNode* y = nil;
   RBNode* x = root;
   while (x != nil)
   {
      y = x;
      if (node->key < x->key)
         x = x->left;
      else
         x = x->right;
   }
   node->parent = y;
   if (y == nil)
      root = node;
   else
      if (node->key < y->key)
         y->left = node;
      else
         y->right = node;
}


void 
RBTree::insert(KEY const& key)
{
   RBNode* p = new RBNode (key, nil);
   rb_insert(p);
}


RBTree::RBNode*
RBTree::remove(RBNode* node)
{
   if (node == nil) 
      return nil;

   _size--;

   RBNode* usuniety = (node->left == nil || node->right == nil) ?
                       node  :  node->next(nil);
   RBNode* galaz = (usuniety->left != nil) ?
                 galaz = usuniety->left  :  usuniety->right;

   galaz->parent = usuniety->parent;
   if (usuniety->parent == nil)
      root = galaz;
   else if (usuniety == usuniety->parent->left)
      usuniety->parent->left = galaz;
   else
      usuniety->parent->right = galaz;
   if (usuniety != node)
      node->key = usuniety->key;
   return usuniety;
}

unsigned int RBTree::RBNode::height(RBTree::RBNode* nil) const
{
   if (this == nil) 
      return 0;
   unsigned int x = left->height(nil);
   unsigned int y = right->height(nil);
   return  (x > y) ? x + 1 : y + 1;
}


unsigned int 
RBTree::RBNode::black_height(RBTree::RBNode* nil) const
{
   if (this == nil) 
      return 1;
   unsigned int x = left->black_height(nil);
   return (color == Red) ? x : x + 1;
}


void 
RBTree::left_rotate(RBTree::RBNode* const node)
{
   assert (node->right != nil);

   RBNode* const syn = node->right;
   node->right = syn->left;
   if (syn->left != nil)
      syn->left->parent = node;
   syn->parent = node->parent;
   if (node->parent == nil)
      root = syn;
   else if (node == node->parent->left)
      node->parent->left = syn;
   else
      node->parent->right = syn;
   syn->left = node;
   node->parent = syn;
}



void RBTree::right_rotate(RBTree::RBNode* node)
{
   assert (node->left != nil);

   RBNode* const syn = node->left;
   node->left = syn->right;
   if (syn->right != nil)
      syn->right->parent = node;
   syn->parent = node->parent;
   if (node->parent == nil)
      root = syn;
   else if (node == node->parent->right)
      node->parent->right = syn;
   else
      node->parent->left = syn;
   syn->right = node;
   node->parent = syn;
}


void RBTree::rb_insert(RBTree::RBNode* node)
{
   insert(node);
   node->color = Red;

   while (node != root && node->parent->color == Red) // 2 czerwone pod rzd
   {
      if (node->parent == node->parent->parent->left) // ojciec z lewego oa
      {
         RBNode* const stryjek = node->parent->parent->right; // stryjek = brat ojca
         if (stryjek->color == Red) // przypadek I: wystarczy przemalowa 
         {
            node->parent->color = Black;
            stryjek->color = Black;
            node->parent->parent->color = Red;
            node = node->parent->parent; // sprawdzamy 2 wezly powyzej
         } 
         else 
         {
            if (node == node->parent->right)// przypadek 2
            {
               node = node->parent;
               left_rotate(node);
            }
            // przypadek 3
            node->parent->color = Black;
            node->parent->parent->color = Red;
            right_rotate(node->parent->parent);
         }
      }
      else
      {
         RBNode* const stryjek = node->parent->parent->left; // stryjek = brat ojca
         if (stryjek->color == Red) // przypadek I: wystarczy przemalowa 
         {
            node->parent->color = Black;
            stryjek->color = Black;
            node->parent->parent->color = Red;
            node = node->parent->parent; // sprawdzamy 2 wezly powyzej
         } 
         else 
         {
            if (node == node->parent->left)// przypadek 2
            {
               node = node->parent;
               right_rotate(node);
            }
            // przypadek 3
            node->parent->color = Black;
            node->parent->parent->color = Red;
            left_rotate(node->parent->parent);
         }
      }
   }
   root->color = Black;
   fix_nil();
#ifndef NDEBUG
   check_it();
#endif
}


RBTree::RBNode* 
RBTree::rb_remove(RBTree::RBNode* node)
{
      if (node == nil) 
      return nil;

   _size--;

   RBNode* usuniety = (node->left == nil || node->right == nil) ?
                       node  :  node->next(nil);
   RBNode* galaz = (usuniety->left != nil) ?
                 galaz = usuniety->left  :  usuniety->right;

   galaz->parent = usuniety->parent;
   if (usuniety->parent == nil)
      root = galaz;
   else if (usuniety == usuniety->parent->left)
      usuniety->parent->left = galaz;
   else
      usuniety->parent->right = galaz;
   if (usuniety != node)
      node->key = usuniety->key;
   if (usuniety->color == Black)
      rb_remove_fixup(galaz);
#ifndef NDEBUG
   check_it();
#endif
   return usuniety;

}

void RBTree::rb_remove_fixup(RBTree::RBNode* node)
{
   while (node != root && node->color == Black)
   {
      if (node == node->parent->left) // node jest "z lewego oa"
      {
         RBNode* brat = node->parent->right;
         if (brat->color == Red)   // przypadek 1: brat jest czerwony
         {

            INFO("fixup: case 1:\n");
            brat->color = Black;
            node->parent->color = Red;
            left_rotate(node->parent);
            brat = node->parent->right; //po obrocie uaktualniamy stryjka 
         }
         if (brat->left->color == Black && brat->right->color == Black)
         {
            // przypadek 2: brat czarny, jego synowie s black-black
            INFO("fixup: case 2:\n");
            brat->color =  Red;
            node = node->parent;
         }
         else 
         {
            if (brat->right->color == Black) 
            {
               // przypadek 3 (brat jest czarny, jego synowie s kolejno red-black)
               INFO("fixup: case 3:\n");
               brat->left->color = Black;
               brat->color = Red;
               right_rotate(brat);
               brat = node->parent->right;
            }
            // przypadek 4 (brat jest czarny, jego synowie kolejno black-red)
            INFO("fixup: case 4:\n");
            brat->color = node->parent->color;
            node->parent->color = Black;
            brat->right->color = Black;
            left_rotate(node->parent);
            node = root; // to przypisanie powoduje przerwanie petli
         }
      }
      else // usuwany wze jest prawym potomkiem
      {
         RBNode* brat = node->parent->left;
         if (brat->color == Red)   // przypadek 1: brat jest czerwony
         {
            INFO("fixup: case 1b:\n");
            brat->color = Black;
            node->parent->color = Red;
            right_rotate(node->parent);
            brat = node->parent->left; //po obrocie uaktualniamy stryjka 
         }
         if (brat->right->color == Black && brat->left->color == Black)
         {
            // przypadek 2: brat czarny, jego synowie s black-black
            INFO("fixup: case 2b:\n");
            brat->color =  Red;
            node = node->parent;
         }
         else // brat ma czerwonego syna
         {
            if (brat->left->color == Black) 
            {
               // przypadek 3 (brat jest czarny, jego synowie s kolejno red-black)
               INFO("fixup: case 3b:\n");
               brat->right->color = Black;
               brat->color = Red;
               left_rotate(brat);
               brat = node->parent->left;
            }
            // przypadek 4 (brat jest czarny, jego synowie kolejno black-red)
            INFO("fixup: case 4b:\n");
            brat->color = node->parent->color;
            node->parent->color = Black;
            brat->left->color = Black;
            right_rotate(node->parent);
            node = root; // to przypisanie powoduje przerwanie petli
         }
      } // konirc kroku petli while
   } // koniec ptli while
   node->color = Black;
   fix_nil();
#ifndef NDEBUG
   check_it();
#endif   
}

void RBTree::fix_nil()
{
   nil->parent = nil;
   nil->left = nil;
   nil->right = nil;
   nil->color = Black;
}


void RBTree::check_it()
{
   assert (nil->key == 0);
   assert (nil->parent == nil);
   assert (nil->left == nil);
   assert (nil->right == nil);
   assert (nil->color == Black);

   assert (root == nil || size() > 0);
   root->check_it(nil);
}


